#!/usr/bin/perl
#
# example:
#
# perl monitor.pl 31000 "./node_control.pl restartone live.conf" nodes.dht

my $debugging = 0;

my $bwthresh = 100;
my $loopwait = 5;
my $printint = 10;
my $getbwint = 10;

use IO::Socket;
STDERR->autoflush(1);
STDOUT->autoflush(1);

if ($#ARGV != 2) {
  die "Usage: prog aqua_port recover_script node_list\n";
}

my $aquaport = shift @ARGV;
my $restart = shift @ARGV;

@hosts;
@host_status;
@host_id;
@host_bw;
@host_bw_t;
$unknown = 0;

while (<>) {
    chomp;
    $_ =~ /^(\S+)\s+(\S+)/;
    
    push @hosts, $1;
    push @host_status, 0;
    push @host_id, 0;
    push @host_bw, 0;
    push @host_bw_t, 0;
}

my @hosts_keys_stored;
my @hosts_bytes_stored;
my @hosts_keys_served;
my @hosts_bytes_served;

for (my $i=0; $i<=$#hosts; $i++) {
  $hosts_keys_stored [$i] = 0;
  $hosts_keys_served [$i] = 0;
  $hosts_bytes_stored [$i] = 0;
  $hosts_bytes_served [$i] = 0;
}

my $getstat_idx = 0;

sub saw_error {
    my ($hostname, $host_idx) = @_;

    $host_status[$host_idx]++;
    if ($host_status[$host_idx] == 10) {
	print "$restarting $hostname\n";
	system("$restart $hostname");
	$host_status[$host_idx] = 0;
    }
    $unknown++;
}

sub getstat_one {
  my $h = $hosts[$getstat_idx];
  # connect to the stats port to see if aqua is up

  my $t = localtime;
  if ($debugging) { print "$t Getting stats for: $h "; }
  my $sock;

  eval {
      local $SIG{ALRM} = sub {die "alarm\n"};
      alarm 10;

      $sock = new IO::Socket::INET 
	  (PeerAddr => $h, PeerPort => $aquaport+2, Proto => 'tcp');
      if ($sock) {
	  while (<$sock>) {
	      chomp;
	      if ($_ =~ /host:([\s\S]+)/) {
		  $host_id[$getstat_idx] = $1;
	      }
	      elsif ($_ =~ /bytes stored:([0-9]*)/) {
		  $hosts_bytes_stored[$getstat_idx] += $1;
	      }
	      elsif ($_ =~ /keys stored:([0-9]*)/) {
		  $hosts_keys_stored[$getstat_idx] += $1;
	      }
	      elsif ($_ =~ /bytes served:([0-9]*)/) {
		  $hosts_bytes_served[$getstat_idx] += $1;
	      }
	      elsif ($_ =~ /keys served:([0-9]*)/) {
		  $hosts_keys_served[$getstat_idx] += $1;
	      }
	  }
	  
	  if ($debugging) { print "(alive, responds to tcp)\n"; }
	  $host_status[$getstat_idx] = 0;
	  close $sock;
      }
      alarm 0;
  };

  # Error conditions
  if (!defined($sock) || $@) {
      my $t = localtime;
      if ($debugging) { print " (may not be up)\n"; }
      if ($@) {
	  print "$t Node $h timed out.  It could be wedged!\n";
      } else {
	  print "$t Node $h not responding.  It could have crashed!\n";
      }
      &saw_error($h, $getstat_idx);
  }
  
  if ($getstat_idx == $#hosts) { $getstat_idx = 0; }
  else { $getstat_idx = $getstat_idx+1; }
}

sub printstats {
  my $t = localtime;
  my $tbsr = 0;
  my $tbst = 0;
  for (my $i=0; $i <= $#hosts; $i++) {
    # my $h = substr ($hosts[$i], 0, index ($hosts[$i],"."));
    my $h = substr ($hosts[$i], 0, 16);
    my $l = length ($host_id[$i]);
    my $c = 8-(40-$l);
    my $n = substr ($host_id[$i], 0, $c);

    print "$t: $h ";
    print "$hosts_keys_stored[$i] $hosts_bytes_stored[$i] ";
    print "$hosts_keys_served[$i] $hosts_bytes_served[$i] ";
    print "$n...\n";
    $tbsr += $hosts_bytes_served[$i];
    $tbst += $hosts_bytes_stored[$i];
  }
  print "$t: total $tbsr bytes served, $tbst bytes stored\n";
}

sub getbw {
  my $t = localtime;
  for (my $i=0; $i <= $#hosts; $i++) {
    my $h = $hosts[$i];
    my $sock = new IO::Socket::INET 
      (PeerAddr => $h, PeerPort => $aquaport-1, Proto => 'tcp');
    if ($sock) {
      while (<$sock>) {
        if ($_ =~ /^([\s\S]+) ([0-9]+.[0-9]+) ([0-9]+.[0-9]+)\n$/) {
	  $host_bw[$i] = $3;
	  $host_bw_t[$i] = $1;
	  if ($3 > $bwthresh) {
            printf ("$1: $h: %.3f KBps (above threshold)\n", $3*1000/8);
	  }
	}
      }
    }
    else {
      $host_bw[$i] = 0;
    }
  }
}

sub printbw {
  my $t = localtime;
  for (my $i=0; $i <= $#hosts; $i++) {
    # my $h = substr ($hosts[$i], 0, index ($hosts[$i],"."));
    my $h = substr ($hosts[$i], 0, 16);
    if ($host_bw[$i] == 0) {
      print "$t: $h has no bw data\n";
    }
    else {
      printf ("$host_bw_t[$i]: $h: %.3f KBps\n", $host_bw[$i]*1000/8);
    }
  }
}

my $t = localtime;
print "$t: reset statistics counters\n";

my $t0 = time ();
my $t1 = time ();

getbw ();
if ($debugging) { printbw (); }

while (1) {
  my $now = time ();
  
  getstat_one ();

  if ($getstat_idx == 0) {
    if ($unknown == 0) { sleep $loopwait; }
    $unknown = 0;
  }
  else {
    select(undef, undef, undef, 0.1);
  }

  if (($now - $t1) > $getbwint) {
    getbw ();
    $t1 = time ();
  }

  if (($now - $t0) > $printint) {
      if ($debugging) { printstats (); }
      if ($debugging) { printbw (); }
      $t0 = time ();
  }
}

