/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#ifndef REPLICA_H
#define REPLICA_H

#include "record.h"
#include "btree.h"
#include "cfsrw.h"

class replica
{
private:
  friend class scanner;

public:
  virtual ~replica () {}
  virtual void exit_cleanup () = 0;
  virtual bigint root () const = 0;
  virtual unsigned nbranches () const = 0;

  virtual str branch_tag (unsigned) const = 0;
  virtual commit_info branch_head_info (unsigned) const = 0;
  virtual bigint branch_head_snapshot (unsigned) const = 0;
  virtual int my_branch () const = 0;

  virtual bool set_branch_head (bigint) = 0;
  virtual int get_branch_index (bigint n) const = 0;
  virtual ptr<btree_srv> btree () const = 0;
  virtual void load_old_commit (bigint c, cbcfsrw) = 0;
  virtual bigint determine_current_branch(str arg_branch, bool ignore_implicit_branches, 
					  str crec_from_info) const = 0;
};

inline bigint
my_branch_head (ptr<replica> r)
{
  int i = r->my_branch ();
  assert (i >= 0);
  return r->branch_head_info (i).rptr;
}

inline bigint
my_branch_head_snapshot (ptr<replica> r)
{
  int i = r->my_branch ();
  assert (i >= 0);
  return r->branch_head_snapshot (i);
}

#endif

