/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "lrucache.h"
#include "pastdbfe.h"
#include "pack.h"
#include "snapshot.h"

ptr<pastdbfe>
create_inmem_db ()
{
  ptr<pastdbfe> db = New refcounted<pastdbfe> ();
  dbOptions opts;
  opts.addOption("opt_flag", DB_CREATE | DB_TRUNCATE);

  // don't supply a path so sleepycat keeps DB in memory
  if (int err = db->opendb(opts)) {
    warn << "cannot open database: " << strerror(err) << " (" << err << ")\n";
    return 0;
  }
  return db;
}

//
// snapshot object methods
//

snapshot::snapshot ()
{
  fmap_db = 0;
  thisaddr = 0;
  ptag = "0";
}

void
snapshot::lookup_fmap (bigint inum, cbsnode_t cb)
{
  ptr<dbrec> q = bigint2dbrec(inum);
  ptr<dbrec> v = fmap_db->lookup(q);
  bigint chash = inum;
  if (v)
    chash = dbrec2bigint (v);
  ns_retrieve_cfs
    (chash,
     wrap (this, &snapshot::lookup_fmap_load_block_cb, cb, inum));
}

void
snapshot::lookup_fmap_load_block_cb (cbsnode_t cb, bigint inum,
                                     ptr<cfs_data> sdata)
{
  if (sdata) {
    snode *e = update_index (inum, sdata);
    cb (e);
  }
  else
    cb ((snode*)0);
}

snode * 
snapshot::update_index (bigint inum, ptr<cfs_data> sdata)
{
  snode *se = New snode (inum, sdata);
  active_snodes.insert (se);
  return se;
}

//
// snapshot_info
//

void
snapshot_info::serialize (strbuf &sb) const
{
  assert (branch.vptr == 0);
  branch.serialize (sb);
  sbigint (sb, prevaddr);
  sbigint (sb, rootdir);
  sstr (sb, btag);
  suint (sb, flags);
  sbigint (sb, supersede);
}

bool
snapshot_info::parse (xin &x)
{
  return commit_info::parse (&x, branch) &&
         x.rbigint (prevaddr) &&
         x.rbigint (rootdir) &&
	 x.rstr (btag) &&
	 x.ruint (flags) &&
	 x.rbigint (supersede);
}

