/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include "stdinc.h"
#include "ns.h"
#include "traverse.h"
#include "view.h"
  
// iterinfo (log_status s, crec r, bigint h, bool f)

traverse::traverse (cbtiter cb, const comvec& start, cbtc tc)
  : _cb (cb), _tc (tc), _start (start)
{
  _work = 0;
  _lerr = 0;
  
  for (unsigned i=0; i < _start.size (); i++) {
    if (_start.rptr (i) != 0 && !_tc (_start.rptr (i))) {
      _todo.push_back (_start.rptr (i));
      _bucket.insert (_start.rptr (i), 0);
    }
  }

  delaycb (0, wrap (this, &traverse::start));
}

void
traverse::start ()
{
  next ();
}

void
traverse::next ()
{
  while (_todo.size () > 0) {
    bigint n = _todo.pop_front ();
    _work++;
    // warn << "fetching " << armor64A (n) << "\n";
    nsobj->retrieve_commit
      (n, wrap (mkref (this), &traverse::load_commit, n));
  }
  if (_work == 0)
    iter ();
}

void
traverse::load_commit (bigint n, crec r, ns_stat status)
{
  _work--;

  if (!r) {
    _lerr++;
    if (_lerr == 1) {
      warn << "cannot load commit " << armor64A (n) << "\n";
      iterinfo info (LOGERR_NOCOMMIT, 0, 0, true);
      _cb (&info);
    }
    return;
  }
 
  unsigned *x = _bucket [n];
  assert (x);
  if (*x == 0) {
    _commits.push_back (r);
    _ids.push_back (n);
    _bucket.remove (n);
    _bucket.insert (n, 1);
  }

  if (r->prev () != 0 && (!_bucket [r->prev ()] && !_tc (r->prev ()))) {
    _todo.push_back (r->prev ());
    _bucket.insert (r->prev (), 0);
  }
  if (r->depends ().rptr != 0 &&
      (!_bucket [r->depends ().rptr] && !_tc (r->depends ().rptr))) {
    _todo.push_back (r->depends ().rptr);
    _bucket.insert (r->depends ().rptr);
  }
  next ();
}

void
traverse::iter ()
{
  _bucket.clear ();

  // warn << "iterating all commit records\n";

  bool changed = true;
  while (changed) {
    changed = false;
    for (unsigned i=0; i<_commits.size (); i++) {
      if (_commits [i] == 0)
	continue;
      crec r = _commits [i];
      bigint n = _ids [i];
      bigint p = r->depends ().rptr;
  
      if (p == 0 || _bucket [p] || _tc (p)) {
	// warn << "changed: " << armor64A (n) << "\n";
	_bucket.insert (n, 0);
	changed = true;
        _commits [i] = 0;
        iterinfo i (LOG_OK, r, n, false);
        _cb (&i);
      }
    }
  }

  // can't order the rest of the commit records, just pass them up
  for (unsigned i=_commits.size (); i > 0; i--) {
    if (_commits [i-1] != 0) {
      crec r = _commits [i-1];
      bigint n = _ids [i-1];
      // warn << "got an unordered commit record " << armor64A (n) << "\n";
      iterinfo i (LOG_OK, r, n, false);
      _cb (&i);
    }
  }

  // warn << "done traversing logs!\n";
  iterinfo i (LOG_OK, 0, 0, true);
  _cb (&i);
  return;
}

