#ifndef _MERKLE_TREE_H_
#define _MERKLE_TREE_H_

#include "async.h"
#include "sha1.h"
#include "merkle_hash.h"
#include "merkle_node.h"
#include "dhash_blockstore.h"

class merkle_tree;

struct merkle_tree_stats {
  uint32 nodes_per_level[merkle_hash::NUM_SLOTS];
  uint32 empty_leaves_per_level[merkle_hash::NUM_SLOTS];
  uint32 leaves_per_level[merkle_hash::NUM_SLOTS];
  uint32 internals_per_level[merkle_hash::NUM_SLOTS];

  uint32 num_nodes;
  uint32 num_leaves;
  uint32 num_empty_leaves;
  uint32 num_internals;
};

struct exclusion_entry
{
  ihash_entry <exclusion_entry> link;
  bigint id;
  exclusion_entry(bigint key) { id = key; }
};

class merkle_tree {
private:

  bool initializing;
  ptr<dbrec> dummydata;

  void rehash (u_int depth, const merkle_hash &key, merkle_node *n);
  void count_blocks (u_int depth, const merkle_hash &key,
		     array<u_int64_t, 64> &nblocks);
  void leaf2internal (u_int depth, const merkle_hash &key, merkle_node *n);
  void remove (u_int depth, block *b, merkle_node *n, bool full_remove);
  int insert (bigint projkey, u_int depth, block *b, merkle_node *n);
  merkle_node *lookup (u_int *depth, u_int max_depth, 
		       const merkle_hash &key, merkle_node *n);
  int insert (bigint projkey, block *b);

public:
  enum { max_depth = merkle_hash::NUM_SLOTS }; // XXX off by one? or two?
  merkle_node root; // ditto
  merkle_tree_stats stats;

  ptr<dbfe> dummydb;
  ptr<dhash_blockstore> storage; // for debug only?

  merkle_tree (ptr<dhash_blockstore> _storage, chordID min, 
	       chordID max, const vec<bigint> &exclude_proj);

  merkle_node *lookup_exact (u_int depth, const merkle_hash &key);
  merkle_node *lookup (u_int depth, const merkle_hash &key);
  merkle_node *lookup (u_int *depth, u_int max_depth, const merkle_hash &key);
  merkle_node *lookup (const merkle_hash &key);
  void clear ();

  void dump ();
  void check_invariants ();
  void compute_stats ();
  void stats_helper (uint depth, merkle_node *n);

};




#endif /* _MERKLE_TREE_H_ */
