/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>
#include "pastwatch.h"
#include "opts.h"
#include "pdb.h"
#include "path.h"
#include "io.h"
#include "cfsw_record.h"

extern ptr<cfsrw> _cfsrw;
extern int check_dirent (str dirpath, dirent *e, bool& ignore_entry);
  
str
get_tag_of_commit (bigint cn)
{
  str stag;
  ptr<btree_srv> srv = _cfsrw->getreplica ()->btree ();
  btree_node x;
  srv->lookup_by_commit (cn, x);
  bigint blah;
  if (!srv->lookup_by_tag (x.proposed_name, blah))
    stag = armor64A (x.commit.rptr);
  else
    stag = x.proposed_name;
  return stag;
}

class status_object : public virtual refcount
{
private:
  vec<str> objects;
  bool isdirop;
  char w;
  cbi cb;

  void
  getattr_cb (str path, bigint fh, unsigned wr, bigint wc, bool sticky,
              cfsrcallp cc, cfsrw_status status)
  {
    if (status)
      cc = 0;

    str wtag;
    ptr<btree_srv> b = _cfsrw->getreplica ()->btree ();
    btree_node x;
    if (b->lookup_by_commit (wc, x)) {
      bigint blah;
      if (!b->lookup_by_tag (x.proposed_name, blah))
        wtag = armor64A (x.commit.rptr);
      else
        wtag = x.proposed_name;
    }

    unsigned rr = 0;
    str rtag;
    str wbtag;
    
    bool same_branch = true;
    if (!cc)
      same_branch = false;

    if (cc) {
      cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
      rr = res->attr->revision;
  
      ptr<btree_srv> b = _cfsrw->getreplica ()->btree ();
      btree_node x;
      if (!b->lookup_by_commit (res->attr->commit, x)) { assert (0); }
      bigint blah;
      if (!b->lookup_by_tag (x.proposed_name, blah))
        rtag = armor64A (x.commit.rptr);
      else
        rtag = x.proposed_name;
      bigint j = 0;
      if (!b->find_join (res->attr->commit, wc, j) ||
          (j != res->attr->commit && j != wc) ||
	  wr > rr)
	same_branch = false;
    }
    
    vec<btree_node> v;
    b->find_leaves (wc, v);
    if (v.size () == 1) {
      unsigned i;
      for (i=0; i<_cfsrw->getreplica ()->nbranches (); i++) {
        if (_cfsrw->getreplica ()->branch_head_info (i).rptr ==
	    v [0].commit.rptr) {
          wbtag = _cfsrw->getreplica ()->branch_tag (i);
          break;
        }
      }
    }

    strbuf out;
    out << "===============================================\n";

    if (fh != 0 && wr != 0) {
      ptr<replica> r = _cfsrw->getreplica ();
      out << "   name: " << path << "\n" 
          << " handle: " << armor64A (fh) << "\n"
	  << " branch: head: "
	  << get_tag_of_commit (my_branch_head (r))
	  << ", tag: "
	  << r->branch_tag (r->my_branch ()) << "\n"
          << "working: r" << wr;

      if (!same_branch) {
	if (wbtag)
	  out << " in branch " << wbtag;
	else
	  out << " in unknown branch";
      }

      if (isdirop) {
	if (w == 'R')
	  out << " (to be removed)";
	else if (w == 'A')
	  out << " (to be relinked)";
	else
	  assert (0);
      }
      // out << ", commit: " << armor64A (wc);
      if (wtag)
	out << ", name: " << wtag;
      if (sticky)
	out << " (sticky)";
      out << "\n";

      if (cc) {
	out << "  repos: r" << rr;
	if (!same_branch)
	  out << " in branch " << r->branch_tag (r->my_branch ());
        // cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
	// out << ", commit: " << armor64A (res->attr->commit);
	out << ", name: " << rtag << "\n";
      }
    }
    else {
      assert (isdirop && w == 'A');
      out << "   name: " << path << "\n" 
          << "working: new file\n";
    }
    out << "\n";
    out.tosuio ()->output (1);
    next ();
  }

  void
  getattr_working_cb (str path, ptr<pfe> p, cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot get attribute of " << path << " in repository\n";
      next ();
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    
    // look for the revision of the file in the latest snapshot on
    // this branch in the repository
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR,
       wrap (mkref (this),
	     &status_object::getattr_cb,
	     path, p->fh, res->attr->revision, res->attr->commit,
	     (p->sticky != 0)));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = p->fh;
    // don't give cfsr_call a snapshot addr, because we want to look
    // at the latest snapshot in the current branch
    cfsr_call (c);
  }

  void
  next ()
  {
    if (objects.empty ()) {
      cb (0);
      return;
    }

    isdirop = false;
    str path = objects.pop_front ();
    
    ptr<pfe> p = pmgr->lookup (path);

    if (!p) {
      str n;
      str d = get_parent_dir (path, n);
      struct dirops dpv;
      if (!get_dirops (get_parent_dir (d), dpv)) {
        for (unsigned i=0; i<dpv.add.size (); i++) {
          if (dpv.add [i]->name == n) {
	    isdirop = true;
	    w = 'A';
	    if (dpv.add [i]->fh == 0) {
              getattr_cb (path, 0, 0, 0, false, 0L, CFSRW_OK);
	      return;
	    }
	    else {
	      p = New refcounted<pfe>;
	      p->name = n;
	      p->fh = dpv.add [i]->fh;
	      p->saddr = 0;
	      break;
	    }
	  }
	}
        for (unsigned i=0; i<dpv.remove.size (); i++) {
          if (dpv.remove [i]->name == n) {
	    isdirop = true;
	    w = 'R';
	    p = New refcounted<pfe>;
	    p->name = n;
	    p->fh = dpv.remove [i]->fh;
	    p->saddr = dpv.remove [i]->saddr;
	    break;
	  }
	}
      }
    }

    if (!p) {
      warn << "cannot lookup " << path << " in local database.\n";
      next ();
      return;
    }
   
    // look for the revision of the file in the working directory
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR,
       wrap (mkref (this), &status_object::getattr_working_cb, path, p));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = p->fh;
    if (p->saddr == 0)
      cfsr_call (c);
    else
      cfsr_call (p->saddr, c);
  }
     
  void
  expand (str path)
  {
    DIR *dir = opendir (path);
    if (dir == 0)
      return;

    struct dirent *e;
    while ((e = readdir (dir))) {
      bool should_ignore = false;
      if (check_dirent (path, e, should_ignore))
	continue;
      if (!should_ignore) {
	str fp;
	if (path == ".")
	  fp = e->d_name;
	else
	  fp = path << "/" << e->d_name;
	objects.push_back (fp);
	if (e->d_type == DT_DIR && arg_recurse)
	  expand (fp);
      }
    }
    closedir (dir);
  }

public:
  status_object (str path, cbi cb)
    : cb (cb)
  {
    objects.push_back (path);
    if (!arg_dironly)
      expand (path);
    next ();
  }
};

void
status (str path, cbi cb)
{
  ptr<status_object> a = New refcounted<status_object> (path, cb);
}


class log_object : public virtual refcount
{
private:
  vec<str> objects;
  bigint root;

  void
  retrieve_cb (bigint fh, bigint ch, unsigned rn, cbi cb, ptr<ns_block> b)
  {
    if (!b) {
      warn << "cannot find the commit record for revision "
	   << rn << " of " << armor64A (fh) << "\n";
      strbuf out;
      out << "\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }

    str s (b->data, b->len);
    crec c = commit_record::alloc (s);

    if (!c) {
      warn << "cannot find the commit record for revision "
	   << rn << " of " << armor64A (fh) << "\n";
      strbuf out;
      out << "\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }

    strbuf out;
    out << "----------------------------\n";
  
    ptr<btree_srv> bt = _cfsrw->getreplica ()->btree ();
    str t = c->proposed_name ();
    bigint blah;
    if (!bt->lookup_by_tag (t, blah))
      t = armor64A (ch);

    out << "revision: " << rn;
    out << ", name: " << t << "\n";
    out << "date: " << myctime ((time_t) c->time ()) << "\n";
    if (c->msg () == "")
      out << "*** empty log message ***\n";
    else
      out << c->msg () << "\n";
    if (rn == 1)
      out << "\n";
    out.tosuio ()->output (1);

    if (rn == 1) {
      next (cb);
      return;
    }
    rn--;

    cfsrcallp cc = New refcounted<cfsrcall>
      (CFSRPROC_FINDVER,
       wrap (mkref (this), &log_object::findver_cb, fh, rn, cb));
    cfsr_findver_arg *arg = static_cast<cfsr_findver_arg *> (cc->arg ());
    arg->object = fh;
    arg->revision = rn;
    cfsr_call (cc);
  }

  void
  getattr_ver_cb (bigint fh, unsigned r, cbi cb,
                  cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot lookup revision " << r << " of " << armor64A (fh) << "\n";
      strbuf out;
      out << "\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    nsobj->retrieve_block
      (res->attr->commit,
       wrap (mkref (this), &log_object::retrieve_cb, fh,
	     res->attr->commit, r, cb));
  }

  void
  findver_cb (bigint fh, unsigned r, cbi cb,
              cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot lookup revision " << r << " of " << armor64A (fh) << "\n";
      strbuf out;
      out << "\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }
    cfsr_findver_res *res = static_cast<cfsr_findver_res *> (cc->res ());
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR,
       wrap (mkref (this), &log_object::getattr_ver_cb, fh, r, cb));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = fh;
    cfsr_call (res->saddr, c);
  }

  void
  getattr_cb (bigint fh, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      warn << "cannot get attribute of " << armor64A (fh) << "\n";
      strbuf out;
      out << "\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }
    cfsr_getattr_res *res = static_cast<cfsr_getattr_res *> (cc->res ());
    unsigned r = res->attr->revision;
    
    strbuf out;
    out << "total revisions in branch: " << res->attr->revision << "\n";
    out.tosuio ()->output (1);

    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_FINDVER,
       wrap (mkref (this), &log_object::findver_cb, fh, r, cb));
    cfsr_findver_arg *arg = static_cast<cfsr_findver_arg *> (c->arg ());
    arg->object = fh;
    arg->revision = r;
    cfsr_call (c);
  }
  
  void
  lookup_cb (str path, cbi cb, cfsrcallp cc, cfsrw_status status)
  {
    if (status) {
      strbuf out;
      out << "name: " << path << "\n" 
          << "not in repository\n\n";
      out.tosuio ()->output (1);
      next (cb);
      return;
    }

    cfsr_lookup_res *res = static_cast<cfsr_lookup_res *> (cc->res ());
    bigint fh = res->object;

    ptr<replica> r = _cfsrw->getreplica ();
    strbuf out;
    out << "name: " << path << "\n" 
        << "handle: " << armor64A (fh) << "\n"
	<< "branch: head: "
	<< get_tag_of_commit (my_branch_head (r))
	<< ", tag:" << r->branch_tag (r->my_branch ()) << "\n";
    out.tosuio ()->output (1);
  
    cfsrcallp c = New refcounted<cfsrcall>
      (CFSRPROC_GETATTR,
       wrap (mkref (this), &log_object::getattr_cb, fh, cb));
    cfsr_getattr_arg *arg = static_cast<cfsr_getattr_arg *> (c->arg ());
    arg->object = fh;
    cfsr_call (c);
  }

  void
  next (cbi cb) 
  {
    if (objects.empty ()) {
      cb (0);
      return;
    }

    str path = objects.pop_front ();
    str cfspath;
    if (get_cfspath (path, cfspath)) {
      cb (-1);
      return;
    }

    if (cfspath == "") { // root dir
      warn << "skipping root dir\n";
      next (cb);
    }
    else { // do lookup to get fh
      strbuf out;
      out << "===============================================\n";
      out.tosuio ()->output (1);

      cfsr_lookup
        (root, cfspath,
         wrap (mkref (this), &log_object::lookup_cb, path, cb));
    }
  }

  void
  expand (str path)
  {
    DIR *dir = opendir (path);
    if (dir == 0)
      return;

    struct dirent *e;
    while ((e = readdir (dir))) {
      bool should_ignore = false;
      if (check_dirent (path, e, should_ignore))
	continue;
      if (!should_ignore) {
	str fp;
	if (path == ".")
	  fp = e->d_name;
	else
	  fp = path << "/" << e->d_name;
	objects.push_back (fp);
	if (e->d_type == DT_DIR && arg_recurse)
	  expand (fp);
      }
    }
    closedir (dir);
  }

public:
  log_object (bigint root, str path, cbi cb)
    : root (root)
  {
    objects.push_back (path);
    if (!arg_dironly)
      expand (path);
    next (cb);
  }
};

void
log (bigint root, str path, cbi cb)
{
  ptr<log_object> a = New refcounted<log_object> (root, path, cb);
}

