#!/usr/bin/perl -w

my $home = $ENV{PASTWATCH_ROOT};
if (!$home || $home eq "") { 
    $home = "$ENV{HOME}/.pastwatch";
}


my $mklog = "pastmklog";
my $viewctl = "pastprojctl";
my $past = "pastwatch";
my $keygen = "pastkeygen";
my $upd = "$home/upd.repositories";

sub usage {
  print "pastctl cmd args\n";
  print "  create [-p projectID] project -- create an empty project\n";
  print "     -p: use given projectID rather than creating a new project\n";
  print "  add [-S timestamp] [-k pubkey] name project -- add a user to an existing project\n";
  print "     -k: use given pubkey for the user\n";
  print "  join [-S timestamp] [-P projectID] project -- create a key pair for a user\n";
  print "  notify rule project -- add notification rule to project\n";
  print "  list project -- list users in a project\n";
  print "  remove project -- removes information about a project\n";
  print "\n";
  print "  args: <project> is the label of a project.\n";
  print "        <projectID> is the hash of the project identifier.\n";
  exit;
};

if ($#ARGV < 0) { usage (); }

my $cmd = $ARGV [0];
shift @ARGV;

if ($cmd eq "remove") {
  if ($#ARGV != 0) {
    usage ();
  }
}
elsif ($cmd eq "create") {
  if (($#ARGV != 0 && $#ARGV != 2) || ($#ARGV == 2 && $ARGV[0] ne "-p")) {
    usage ();
  }
}
elsif ($cmd eq "add") {
    if (($#ARGV != 1 && $#ARGV != 3 && $#ARGV != 5) || 
	($#ARGV == 3 && ($ARGV[0] ne "-k" && $ARGV[0] ne "-S")) ||
	($#ARGV == 5 && ($ARGV[0] ne "-S" && $ARGV[2] ne "-k"))) {
    usage ();
  }
}
elsif ($cmd eq "join") {
  if (($#ARGV != 0 && $#ARGV != 2 && $#ARGV != 4)) {
    usage ();
  }
}
elsif ($cmd eq "list") {
  if ($#ARGV != 0) {
    usage ();
  }
}
elsif ($cmd eq "notify") {
  if ($#ARGV != 1) {
    usage ();
  }
}
else { usage (); }


# Parse the last argument and prematurely set $mydir
my $label = $ARGV[$#ARGV];
my $mydir = "$home/$label";

sub getview {
  my ($dir) = @_;
  if (-l "$dir/project.active") {
    my $s = readlink "$mydir/project.active";
    if ($s =~ /^project:([\s\S]+)$/) { return $1; }
    else { die "$mydir/project.active -- corrupted\n"; }
  }
  else {
    die "cannot find $dir/project.active\n";
  }
}

sub create_user {
  my ($nview, $timestamp) = @_;

  my $r = system ("$keygen -KP -l $label /tmp/user.key.$$");
  if ($r) {
    unlink "/tmp/user.key.$$";
    die "cannot create user's key file\n";
  }

  my $ts = "";
  if ($timestamp && $timestamp ne "") { $ts = "-S \"$timestamp\" "; }

  my $list = `$viewctl -p $mydir/project.active 2>&1`;
  if ($list =~ /cannot load project/) { die "Could not load project"; }

  my @list = split /\n/, $list;
  my $command;
  if (scalar(@list) < 1) {
      $command = "$mklog $ts -c -s /tmp/user.key.$$ $nview 1>/tmp/mklog.$$";
  } else {
      $command = "$mklog $ts -s /tmp/user.key.$$ $nview 1>/tmp/mklog.$$";
  }

  $r = system ($command);
  if ($r) {
    unlink "/tmp/user.key.$$";
    unlink "/tmp/mklog.$$";
    die "cannot create user's log pointer\n";
  }

  my $nkey = `cat /tmp/mklog.$$`;
  unlink "/tmp/mklog.$$";
  chomp $nkey;

  $r = system ("mv /tmp/user.key.$$ $mydir/key.$nkey");
  if ($r) {
    unlink "/tmp/user.key.$$";
    die "cannot move user's key file into $mydir\n";
  }

  return $nkey;
}

sub create_proj_dir {
    my ($home, $projname, $projid) = @_;

    print "mkdir $home $projname\n";

    mkdir ($home,0755);
    if (!mkdir ("$home/$projname",0755)) { die "cannot create $home/$projname\n"; }

    my $r = system ("cd $home/$projname; ln -sf project:$projid project.active");
    if ($r) { die "cannot create symlink $home/$projname/project.active\n"; }
}

#
# start running the commands
#

my $r;

if ($cmd eq "remove") {
  my $nview = getview ($mydir);
  open (F0, $upd) || die "can't remove project from $upd\n";
  open (F1, ">$upd.$$") || die "can't open $upd.$$\n";
  while (<F0>) {
    chomp;
    if ($_ ne $nview) { print F1 "$_\n"; }
  }
  close F1;
  close F0;
  rename ("$upd.$$", "$upd");
  system ("rm -rf $mydir");
}

elsif ($cmd eq "create") {
  if (-x $mydir) { die "$mydir exists: remove it first\n"; }
  my $nview;

  mkdir ($home,0755);
  if (!mkdir ($mydir,0755)) { die "cannot create $mydir\n"; }

  if ($#ARGV == 2) {
    $nview = $ARGV[1];
  }
  else {
    $r = system ("$keygen -KP -l $label /tmp/project.key.$$");
    if ($r) {
      unlink "/tmp/project.key.$$";
      die "cannot create project's key file\n";
    }

    $r = system ("$viewctl -s /tmp/project.key.$$ 1>/tmp/projectctl.$$");
    if ($r) {
      unlink "/tmp/project.key.$$";
      unlink "/tmp/projectctl.$$";
      die "cannot create project\n";
    }

    $nview = `cat /tmp/projectctl.$$`;
    unlink "/tmp/projectctl.$$";
    chomp $nview;

    $r = system ("mv /tmp/project.key.$$ $mydir/project.key");
    if ($r) {
      unlink "/tmp/project.key.$$";
      die "cannot move project's key file into $mydir\n";
    }
  }
  
  $r = system ("cd $mydir; ln -sf project:$nview project.active");
  if ($r) { die "cannot create symlink $mydir/project.active\n"; }
  print "new project ID: $nview\n";

  open (F0, ">>$upd")
    || die "can't update $upd - disconnected commits won't propagate\n";
  print F0 "$nview\n";
  close F0;
}

elsif ($cmd eq "add") {
  if (!(-e "$mydir/project.key")) { die "cannot find $mydir\n"; }
  my $nview = getview ($mydir);
  my $name = $ARGV[$#ARGV-1];
  my $nkey;
  my $timestamp;

  # parse arguments
  my $i = 0;
  while ($i < $#ARGV) {
      if ($ARGV[$i] eq "-k") {
	  $nkey = $ARGV[$i+1];
	  $i += 2;
      } elsif ($ARGV[$i] eq "-S") {
	  $timestamp = $ARGV[$i+1];
	  $i += 2;
      } else {
	  $i ++;
      }
  }

  if (!$nkey || $nkey eq "")  { 
      $nkey = create_user ($nview, $timestamp); 
  }

  $r = system
    ("$viewctl -s $mydir/project.key -n $name -a $nkey $nview 1> /dev/null");
  if ($r) { die "cannot add user to project\n"; }
  print "new user: $name, $nkey\n";
}

elsif ($cmd eq "join") {
    my $timestamp = "";
    my $projectid = "";

    # Parse args
    for(my $i=0; $i<scalar(@ARGV); $i++) {

	if($ARGV[$i] eq "-S") {
	    if ($i < $#ARGV) {
		$timestamp = $ARGV[++$i];
	    } else {
		die "Error with -S argument.";
		&usage();
	    }
	} elsif ($ARGV[$i] eq "-P") {
	    if ($i < $#ARGV) {
		$projectid = $ARGV[++$i];
	    } else {
		die "Error with -P argument.";
		&usage();
	    }
	} elsif ($i == $#ARGV ) {
	} else {
	    &usage();
	}
    }
    
    # If the user specified a key, then create a new directory
    if ($projectid ne "") {
	&create_proj_dir($home, $label, $projectid);
    }
    
    my $nview = getview ($mydir);
    my $nkey = create_user ($nview, $timestamp);

  print "pass $nkey and a user name to project administrator\n";
}

elsif ($cmd eq "notify") {
  if (!(-e "$mydir/project.key")) { die "cannot find $mydir\n"; }
  my $nview = getview ($mydir);
  my $rule = $ARGV[0];

  $r = system ("$viewctl -r \"$rule\" -s $mydir/project.key $nview 1>/dev/null");
  if ($r) {
    die "cannot add notification rule to project\n";
  }
}

elsif ($cmd eq "list") {
  my $nview = getview ($mydir);
  my $list = `$viewctl -p $mydir/project.active`;
  my @list = split /\n/, $list;
  my $line;
  
  print "project $label: $nview\n";
  foreach $line (@list) {
    if ($line =~ /user: ([\s\S]+) \(([\s\S]+)\)/) {
      if (-r "$mydir/key.$2") { print "  * user: $1 ($2)\n"; }
      else { print "    user: $1 ($2)\n"; }
    }
    else { print "  $line\n"; }
  }
}


