/* 
 * The code below is copied from the SFS software on June 30th, 2003.
 * It was part of the sfskey program. see http://www.fs.net/
 */
/*
 *  Copyright (C) 2002-2004  Massachusetts Institute of Technology
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 *  USA
 */
/*
 *
 * Copyright (C) 1999 Michael Kaminsky (kaminsky@lcs.mit.edu)
 * Copyright (C) 1998, 1999 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "parseopt.h"
#include "crypt.h"
#include "srp.h"
#include "sfskeymgr.h"

#define RANDOM_SEED "~/.sfs/random_seed"

void
usage ()
{
  warn << "options are: [-KP] [-b nbits] [-c cost] [-l label] [keyname]\n";
  exit (1);
}

void
sfskey_gen (int argc, char **argv)
{   
  u_int nbits = 0, cost = 0;
  str keyname, keylabel, user;
  u_int32_t opts = KM_FGEN;
  
  int ch;
  while ((ch = getopt (argc, argv, "b:c:l:u:KP")) != -1)
    switch (ch) {
    case 'b':
      if (!convertint (optarg, &nbits))
        usage ();
      break;
    case 'c':
      if (!convertint (optarg, &cost))
        usage ();
      if (cost > sfs_maxpwdcost)
        fatal ("password cost must be far less than %d\n", sfs_mindlogsize);
      break;
    case 'l':
      keylabel = optarg;
      break;
    case 'u':
      user = optarg;
      break;
    case 'K':
      opts |= KM_NOKBD;
      break;
    case 'P':
      opts |= KM_NOPWD;
      break;
    default:
      usage ();
      break;
    }

  if (optind + 1 == argc)
    keyname = argv[optind];
  else if (optind != argc)
    usage ();

  str err;
  sfskeymgr kmgr (user);
  sfskeyinfo *ki;
  sfskey *k;
  if (!(ki = kmgr.getkeyinfo (keyname, opts)))
    fatal << "No suitable key target found\n";
  ki->keylabel = keylabel;
  if (!(k = kmgr.fetch_or_gen (ki, &err, nbits, cost, opts)))
    fatal << err << "\n";
  if (!kmgr.save (k, ki, opts))
    fatal << "Not all key save operations succeeded\n";
  exit (0);
}

int
main (int argc, char **argv)
{
  sfskey_gen (argc, argv);
  return 0;
}

