#include "blacklist.h"
#include "admcontrol.h"
#include <misc_utils.h>
#include <serial.h>
#include <rxx.h>

blacklist::blacklist() {}
blacklist::~blacklist() {}

static rxx projkey_rxx ("(^[0-9a-zA-Z+-]+)$");
static rxx blacklist_format ("^([\\s\\S]+?)\n(Version\\s+(\\d+)\\s*\n([\\s\\S]*))$"); 
void
blacklist::init(str dirname, str name, int port)
{
  vec<str> lines;
  int r, i;
  sfs_sig2 sig;
  assert (!node_list[name_port2str(name, port)]);
  node_list.insert( New node_entry(name, port) );

  str filename = strbuf() << dirname << "/" << name << ":" << port;
  str contents = file2str(filename);

  warn << "Blacklist for: " << name << ":" << port << "\n";
  if (!contents)
    return;

  if (!blacklist_format.match(contents)) {
    warn << "Could not parse blacklist for " << name << ":" << port << "\n";
    return;
  }

  if (!str2sfs_sig2(blacklist_format[1], &sig)) {
    warn << "could not parse signature for " << name << ":" << port << "\n";
    return;
  }

  struct hostentry *p = hostlist::only().find(name, port);
  assert(p);
  
  if (!p->verifier->verify(sig, blacklist_format[2])) {
    warn << "could not verify blacklist for " << name << ":" << port << "\n";
    return;
  }

  node_list[name_port2str(name,port)]->version = atoi(blacklist_format[3]);
  r = split(&lines, "\n", blacklist_format[4]);
  
  for(i=0; i<r; i++) {
    if (projkey_rxx.match (lines[i])) {
      warn << "  Unwilling to store: " << projkey_rxx[1] << "\n";
      node_list[name_port2str(name,port)]->blist.insert 
	( New project_entry ( dearmor64A_bigint(projkey_rxx[1])) );
      
    } else
      warn << "  error parsing blocklist line " << 1+i << "\n";
  }
}

bool 
blacklist::deny_project(str name, int port, bigint projkey) 
{
  if (blacklist_dir == "")
    return false;

  return node_list[name_port2str(name,port)]->blist[projkey];
}

bool 
blacklist::deny_project(const net_address &addr, bigint projkey) 
{
  return deny_project(addr.hostname, addr.port, projkey);
}

void
blacklist::append_blacklist(str name, int port, vec<bigint> &list)
{
  const project_entry *proj = node_list[name_port2str(name,port)]->blist.first();
  while(proj) {
    list.push_back(proj->id);
    proj = node_list[name_port2str(name,port)]->blist.next(proj);
  }
}

void
blacklist::append_blacklist(const net_address &addr, vec<bigint> &list)
{
  append_blacklist(addr.hostname, addr.port, list);
}

int
blacklist::set_dir(str dirname)
{
  blacklist_dir = dirname;
  return 0;
}

void
blacklist::read_all() 
{
  struct hostentry *p = hostlist::only().first();
    
  while(p) {
    init(blacklist_dir, p->name, p->port);
    p = hostlist::only().next(p);
  }
}

int
blacklist::getversion(str name, int port)
{
  if (blacklist_dir == "")
    return 0;

  return node_list[name_port2str(name,port)]->version;
  
}


blacklist &
blacklist::only ()
{
  static ptr<blacklist> instance = NULL;
  if (instance == NULL)
    instance = New refcounted<blacklist> ();
  return *instance;
}

str name_port2str(str name, int port) {
  strbuf sb = strbuf() << name << ":" << port;
  return str(sb);
}
